#version 430

layout(local_size_x = 128, local_size_y = 1, local_size_z = 1) in;

struct Data {
	vec4 values;
};

layout(std430, binding = 0) buffer data1
{
	Data positions[];
};

layout(std430, binding = 1) buffer data2
{
	Data uvs[];
};

layout(std430, binding = 2) buffer data3
{
	Data normals[];
};

layout(std430, binding = 3) buffer data4
{
	Data tangents[];
};

layout(std430, binding = 4) buffer data5
{
	unsigned int indices[];
};

layout(std430, binding = 5) buffer data6
{
	float heightField[];
};

uniform int uMode;
uniform int uGridSize;
uniform vec3 uPosition;
uniform float uSize;
uniform sampler2D uHeightMap;
uniform float uHeightScale;

uint getInvocationIndex() {
	return gl_GlobalInvocationID.x +
		gl_GlobalInvocationID.y *
		gl_NumWorkGroups.x *
		gl_WorkGroupSize.x;
}

void main() {
	uint idx = getInvocationIndex();	
	uint mode = uint(uMode);
	if (mode == 0) {
		// positions and uvs
		uint g = uint(uGridSize);
		uint g2 = g * g;
		if (idx < g2) {
			float x = float(idx % g);
			float y = float(idx / g);
			vec2 uv = vec2(
				x / float(g - 1),
				1.0 - (y / float(g - 1))
			);
			vec2 minPosition = uPosition.xz - vec2(uSize / 2.0, uSize / 2.0);
			float gridScale = uSize / float(g - 1);
			float h = pow(clamp(texture2D(uHeightMap, uv).r, 0.0, 1.0), 1.5);
			float height = uPosition.y + h * uHeightScale;
			positions[idx].values = vec4(
				minPosition.x + x * gridScale,
				height,
				minPosition.y + y * gridScale,
				0
			);	
			uvs[idx].values = vec4(
				uv,
				0,
				0
			);
			heightField[idx] = height;
		}
	}
	else if (mode == 1) {
		// indices
		uint g = uint(uGridSize - 1);
		uint g2 = g * g;
		if (idx < g2) {

			uint xW = idx % g;
			uint yN = idx / g;
			uint xE = xW + 1;
			uint yS = yN + 1;

			uint iWN = yN * (g + 1) + xW;
			uint iEN = yN * (g + 1) + xE;
			uint iWS = yS * (g + 1) + xW;
			uint iES = yS * (g + 1) + xE;

			indices[6 * idx + 0] = iWN;
			indices[6 * idx + 1] = iWS;
			indices[6 * idx + 2] = iEN;
			indices[6 * idx + 3] = iEN;
			indices[6 * idx + 4] = iWS;
			indices[6 * idx + 5] = iES;
		}
	}
	else if (mode == 2) {
		// normals
		uint g = uint(uGridSize);
		uint g2 = g * g;
		if (idx < g2) {			
			uint xi = idx % g;
			uint yi = idx / g;
			float xNext = positions[yi * g + min(xi + 1, g - 1)].values.y;
			float xPrev = positions[yi * g + (xi > 0 ? xi - 1 : xi)].values.y;
			float yNext = positions[min(yi + 1, g - 1) * g + xi].values.y;
			float yPrev = positions[(yi > 0 ? yi - 1 : yi) * g + xi].values.y;

			normals[idx].values = vec4(
				normalize(vec3(
					xPrev - xNext,
					2.0, 
					yPrev - yNext
				)),
				0
			);
		}
	}
	else if (mode == 3) {
		// tangents
		uint g = uint(uGridSize - 1);
		uint g2 = g * g * 2; // triangles
		if (idx < g2) {
			uint i0 = indices[3 * idx + 0];
			uint i1 = indices[3 * idx + 1];
			uint i2 = indices[3 * idx + 2];

			vec3 position0 = positions[i0].values.xyz;
			vec3 position1 = positions[i1].values.xyz;
			vec3 position2 = positions[i2].values.xyz;

			vec3 position01 = position1 - position0;
			vec3 position02 = position2 - position0;

			vec2 uv0 = uvs[i0].values.xy;
			vec2 uv1 = uvs[i1].values.xy;
			vec2 uv2 = uvs[i2].values.xy;

			vec2 uv01 = uv1 - uv0;
			vec2 uv02 = uv2 - uv0;

			float r = 1.0 / (uv01.x * uv02.y - uv01.y * uv02.x);
			vec4 tangent = vec4(
				normalize(
					(position01 * uv02.y - position02 * uv01.y) * r
				),
				0
			);
			tangents[i0].values = tangent;
			tangents[i1].values = tangent;
			tangents[i2].values = tangent;
		}
	}
}