#version 330

layout (location = 0) out vec4 outColor;
layout (location = 1) out vec4 outDepth;

in ES_OUT{
	vec3 vsPosition;
	vec2 uv;
	vec3 tsToLight;
	vec3 tsToCamera;
	vec4 smPosition;
} fs_in;

// material
struct Material {
	sampler2D maskMap;

	sampler2D displacementMap;
	float displacementStrength;

	float tilingR;
	vec3 ambientR;
	sampler2D diffuseMapR;
	vec3 specularR;
	float shininessR;
	sampler2D normalMapR;
	int useNormalMapR;

	float tilingG;
	vec3 ambientG;
	sampler2D diffuseMapG;
	vec3 specularG;
	float shininessG;
	sampler2D normalMapG;
	int useNormalMapG;

	float tilingB;
	vec3 ambientB;
	sampler2D diffuseMapB;
	vec3 specularB;
	float shininessB;
	sampler2D normalMapB;
	int useNormalMapB;
};
uniform Material uMaterial;

uniform sampler2D uShadowMap;

uniform int uNormalMapping;

// fog
struct Fog {
	float density;
	float gradient;
	vec3 color;
};
uniform Fog uFog;

// light
struct DirectionalLight {
	vec3 direction;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};
uniform DirectionalLight uLight0;

void main() {
	vec3 mask = texture2D(uMaterial.maskMap, fs_in.uv).rgb;
	float maskMax = clamp(mask.r + mask.g + mask.b, 0.01, 3.0);
	float r = clamp(mask.r / maskMax, 0, 1);
	float g = clamp(mask.g / maskMax, 0, 1);
	float b = clamp(mask.b / maskMax, 0, 1);

	vec3 diffuseColorR = texture2D(uMaterial.diffuseMapR, fs_in.uv * uMaterial.tilingR).rgb;
	vec3 diffuseColorG = texture2D(uMaterial.diffuseMapG, fs_in.uv * uMaterial.tilingG).rgb;
	vec3 diffuseColorB = texture2D(uMaterial.diffuseMapB, fs_in.uv * uMaterial.tilingB).rgb;
		
	vec3 diffuseColor =
		diffuseColorR * r +
		diffuseColorG * g +
		diffuseColorB * b;
	vec3 materialAmbient =
		uMaterial.ambientR * r +
		uMaterial.ambientG * g +
		uMaterial.ambientB * b;
	vec3 materialSpecular =
		uMaterial.specularR * r +
		uMaterial.specularG * g +
		uMaterial.specularB * b;
	float materialShininess =
		uMaterial.shininessR * r +
		uMaterial.shininessG * g +
		uMaterial.shininessB * b;

	vec3 tsDefaultNormal = vec3(0, 0, 1);
	vec3 tsNormalR = (uNormalMapping > 0 && uMaterial.useNormalMapR > 0) ?
		normalize(texture2D(uMaterial.normalMapR, fs_in.uv * uMaterial.tilingR).rgb * 2.0 - 1.0) : tsDefaultNormal;
	vec3 tsNormalG = (uNormalMapping > 0 && uMaterial.useNormalMapG > 0) ?
		normalize(texture2D(uMaterial.normalMapG, fs_in.uv * uMaterial.tilingG).rgb * 2.0 - 1.0) : tsDefaultNormal;
	vec3 tsNormalB = (uNormalMapping > 0 && uMaterial.useNormalMapB > 0) ?
		normalize(texture2D(uMaterial.normalMapB, fs_in.uv * uMaterial.tilingB).rgb * 2.0 - 1.0) : tsDefaultNormal;
	vec3 tsNormal = normalize(r * tsNormalR + g * tsNormalG + b * tsNormalB);

	float cameraDistance = length(fs_in.vsPosition);
	vec3 tsToLight = normalize(fs_in.tsToLight);
	vec3 tsToCamera = normalize(fs_in.tsToCamera);

	vec3 ambient = uLight0.ambient * materialAmbient * diffuseColor;

	float diffuseFactor = clamp(dot(tsNormal, tsToLight), 0, 1);
	vec3 diffuse = uLight0.diffuse * diffuseFactor * diffuseColor;

	vec3 reflected = reflect(-tsToLight, tsNormal);
	float specularFactor = pow(clamp(dot(tsToCamera, reflected), 0, 1), materialShininess);
	vec3 specular = uLight0.specular * specularFactor * materialSpecular;

	// shadow mapping
	float shadow = 1.0;
	float shadowIn = 50.0;
	float shadowFull = 20.0;
	if (cameraDistance <= shadowIn) {
		float bias = 0;
		vec2 texelSize = 1.0 / textureSize(uShadowMap, 0);
		vec3 shadowCoords = fs_in.smPosition.xyz * 0.5 + 0.5;
		int sampleStep = 3;
		float sampleCount = pow(float(sampleStep) * 2.0 + 1, 2.0);
		if (shadowCoords.x >= 0 && shadowCoords.x <= 1 && shadowCoords.y >= 0 && shadowCoords.y <= 1)
			for (int y = -sampleStep; y <= sampleStep; y++)
				for (int x = -sampleStep; x <= sampleStep; x++) {
					float pcfDepth = texture2D(uShadowMap, shadowCoords.xy + vec2(x, y) * texelSize).r;
					shadow -= shadowCoords.z - bias > pcfDepth ? 1.0 / sampleCount : 0.0;
				}

		// interpolate
		float a = clamp((shadowIn - cameraDistance) / (shadowIn - shadowFull), 0, 1);
		shadow = (1.0 - a) + a * shadow;
	}

	vec3 sceneColor = ambient + (diffuse + specular) * shadow;
	float v = clamp(
		exp(-pow(cameraDistance * uFog.density, uFog.gradient)),
		0, 1);
	outColor = vec4(sceneColor * v + uFog.color * (1.0 - v), 1.0);
	outDepth = vec4(cameraDistance, 0, 0, 1);
}