#version 410

layout(vertices = 3) out;

in VS_OUT {
	vec3 wsPosition;
	vec3 vsPosition;
	vec2 uv;
	vec3 wsNormal;
	vec3 wsTangent;
	float tessellationRatio;
} cs_in[];

out CS_OUT{
	vec3 wsPosition;
	vec3 vsPosition;
	vec2 uv;
	vec3 wsNormal;
	vec3 wsTangent;
	float tessellationRatio;
} es_in[];

uniform mat4 uProjectionMatrix;

uniform int uTesselation;

// view frustum culling
bool outsideFrustum(vec3 p1, vec3 p2, vec3 p3) {
	float s = 1.2;
	return 
		((p1.x < -s && p2.x < -s && p3.x < -s) ||
		(p1.x > s && p2.x > s && p3.x > s) ||
		(p1.y < -s && p2.y < -s && p3.y < -s) ||
		(p1.y > s && p2.y > s && p3.y > s) ||
		(p1.z < -s && p2.z < -s && p3.z < -s) ||
		(p1.z > s && p2.z > s && p3.z > s));
}

// backfaceculling
bool isFrontface(vec3 p1, vec3 p2, vec3 p3) {
	vec3 toEye = normalize(-((p1 + p2 + p3) / 3.0));
	vec3 normal = normalize(cross(p2 - p1, p3 - p1));
	return dot(normal, toEye) >= -0.1;
}

// tessellation level
float getTessLevel(float d) {
	return d < 25 ? 15 : 1;
}

void main() {
	if (gl_InvocationID == 0) {
		bool culled = false;
		// frustum culling
		vec4 csPosition0 = uProjectionMatrix * vec4(cs_in[0].vsPosition, 1.0);
		vec4 csPosition1 = uProjectionMatrix * vec4(cs_in[1].vsPosition, 1.0);
		vec4 csPosition2 = uProjectionMatrix * vec4(cs_in[2].vsPosition, 1.0);
		bool shouldTessellate = 
			cs_in[0].tessellationRatio > 0 || 
			cs_in[1].tessellationRatio > 0 || 
			cs_in[2].tessellationRatio > 0;
		if (outsideFrustum(csPosition0.xyz / csPosition0.w, csPosition1.xyz / csPosition1.w, csPosition2.xyz / csPosition2.w))
			culled = true;

		// backface culling
		if (!culled && !isFrontface(cs_in[0].vsPosition, cs_in[1].vsPosition, cs_in[2].vsPosition))
			culled = true;

		if (!culled) {
			// set adaptive tesslevel
			if (uTesselation > 0 && shouldTessellate) {

				float d0 = length(cs_in[0].vsPosition);
				float d1 = length(cs_in[1].vsPosition);
				float d2 = length(cs_in[2].vsPosition);

				gl_TessLevelInner[0] = getTessLevel((d0 + d1 + d2) / 3.0);
				gl_TessLevelOuter[0] = getTessLevel((d1 + d2) / 2.0);
				gl_TessLevelOuter[1] = getTessLevel((d2 + d0) / 2.0);
				gl_TessLevelOuter[2] = getTessLevel((d0 + d1) / 2.0);
			}
			else {
				gl_TessLevelInner[0] = 1;
				gl_TessLevelOuter[0] = 1;
				gl_TessLevelOuter[1] = 1;
				gl_TessLevelOuter[2] = 1;
			}
		}
		else {
			gl_TessLevelInner[0] = 0;
			gl_TessLevelOuter[0] = 0;
			gl_TessLevelOuter[1] = 0;
			gl_TessLevelOuter[2] = 0;
		}
	}

	es_in[gl_InvocationID].wsPosition = cs_in[gl_InvocationID].wsPosition;
	es_in[gl_InvocationID].vsPosition = cs_in[gl_InvocationID].vsPosition;
	es_in[gl_InvocationID].uv = cs_in[gl_InvocationID].uv;
	es_in[gl_InvocationID].wsNormal = normalize(cs_in[gl_InvocationID].wsNormal);
	es_in[gl_InvocationID].wsTangent = normalize(cs_in[gl_InvocationID].wsTangent);
	es_in[gl_InvocationID].tessellationRatio = cs_in[gl_InvocationID].tessellationRatio;
}