#version 330

layout (location = 0) out vec4 outColor;
layout (location = 1) out vec4 outDepth;

in VS_OUT {
	vec3 wsPosition;
	vec3 vsPosition;
	vec4 csPosition;
	vec2 uv;
	vec3 tsToLight;
	vec3 tsToCamera;
} fs_in;

// material
struct Material {
	sampler2D refractionMap;
	sampler2D depthMap;
	sampler2D reflectionMap;
	sampler2D dudvMap;
	sampler2D normalMap;
};
uniform Material uMaterial;

// fog
struct Fog {
	float density;
	float gradient;
	vec3 color;
};
uniform Fog uFog;

// light
struct DirectionalLight {
	vec3 direction;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};
uniform DirectionalLight uLight0;

uniform vec3 uCameraPosition;
uniform float t;

uniform int uNormalMapping;

const float waveStrength = 0.04;
const float shininess = 20.0;
const float reflectivity = 0.5;

void main() {
	vec3 waterColor = uFog.color;

	vec2 ndc = (fs_in.csPosition.xy / fs_in.csPosition.w) / 2.0 + 0.5;
	vec2 refractionCoords = vec2(ndc.x, -1.0 + ndc.y);
	vec2 reflectionCoords = vec2(ndc.x, 1.0 - ndc.y);

	float sceneVsDepth = texture2D(uMaterial.depthMap, refractionCoords).r;
	float vsDepth = length(fs_in.vsPosition);
	float waterDepth = sceneVsDepth - vsDepth;
	float waterAlpha = clamp(waterDepth * 5.0, 0, 1);

	vec2 distCoords = texture2D(uMaterial.dudvMap, vec2(fs_in.uv.x + t, fs_in.uv.y)).rg * 0.1;
	distCoords = fs_in.uv + vec2(distCoords.x, distCoords.y + t);
	vec2 totalDist = (texture2D(uMaterial.dudvMap, distCoords).rg * 2.0 - 1.0) * waveStrength;
	refractionCoords.x = clamp(refractionCoords.x + totalDist.x, 0.001, 0.999);
	refractionCoords.y = clamp(refractionCoords.y + totalDist.y, -0.999 , -0.001);
	reflectionCoords = clamp(reflectionCoords + totalDist, 0.001, 0.999);
	vec3 refractionColor = texture2D(uMaterial.refractionMap, refractionCoords).rgb;
	vec3 reflectionColor = texture2D(uMaterial.reflectionMap, reflectionCoords).rgb;

	vec3 wsToCamera = normalize(uCameraPosition - fs_in.wsPosition);
	vec3 tsToCamera = normalize(fs_in.tsToCamera);
	vec3 tsToLight = normalize(fs_in.tsToLight);
	vec3 tsNormal = uNormalMapping > 0 ? 
		normalize(texture2D(uMaterial.normalMap, distCoords).rgb * 2.0 - 1.0) : vec3(0, 0, 1);

	vec3 diffuseColor = mix(
		reflectionColor, 
		refractionColor, 
		clamp(dot(wsToCamera, vec3(0, 1, 0)), 0, 1));
	diffuseColor = mix(waterColor, diffuseColor, 0.3);

	vec3 ambient = uLight0.ambient * vec3(0.4, 0.4, 0.4) * diffuseColor;

	float diffuseFactor = clamp(dot(tsNormal, tsToLight), 0, 1);
	vec3 diffuse = uLight0.diffuse * diffuseFactor * diffuseColor;

	vec3 reflected = reflect(-tsToLight, tsNormal);
	float specularFactor = pow(clamp(dot(tsToCamera, reflected), 0, 1), shininess);
	vec3 specular = uLight0.specular * specularFactor * reflectivity;

	vec4 sceneColor = vec4(ambient + diffuse + specular, waterAlpha);

	float v = clamp(
		exp(-pow(length(fs_in.vsPosition) * uFog.density, uFog.gradient)),
		0, 1);
	outColor = sceneColor * v + vec4(uFog.color, 1.0) * (1.0 - v);
	outDepth = vec4(length(fs_in.vsPosition), 0, 0, 1);
}